module Buffering
  include Enumerable
  attr_accessor :sync
  BLOCK_SIZE = 4096

  private

  def fill_rbuff
    @rbuffer = "" unless @rbuffer
    begin
      IO.select([self.to_io], nil, nil) if self.respond_to?(:to_io)
      @rbuffer << self.primitive_read(BLOCK_SIZE)
    rescue EOFError
      @eof = true
    end
  end

  def consume_rbuff(size=nil)
    if @rbuffer.size == 0
      @eof = nil
      nil
    else
      size = @rbuffer.size unless size
      ret = @rbuffer[0, size]
      @rbuffer[0, size] = ""
      ret
    end
  end

  public

  def read(size=nil)
    fill_rbuff unless @rbuffer
    until @eof
      break if size && size <= @rbuffer.size
      fill_rbuff
    end
    consume_rbuff(size)
  end

  def gets(eol=$/)
    fill_rbuff unless @rbuffer
    idx = @rbuffer.index(eol)
    until @eof
      break if idx
      fill_rbuff
      idx = @rbuffer.index(eol)
    end
    if eol.is_a?(Regexp)
      size = idx ? idx+$&.size : nil
    else
      size = idx ? idx+eol.size : nil
    end
    consume_rbuff(size)
  end

  def getc
    c = read(1)
    c ? c.to_i : nil
  end

  def ungetc(c)
    @buffer[0,0] = c.chr
  end

  def eof?
    @eof && @rbuffer.size == 0
  end

  def each
    while line = self.gets
      yield line
    end
  end

  def readlines
    ary = []
    while line = self.gets
      ary << line
    end
    ary
  end

  private

  def consume_wbuf(len)
    remainder = len
    written = 0
    while remainder > 0
      IO.select(nil, [self.to_io], nil) if self.respond_to?(:to_io)
      nwrite = self.primitive_write(@wbuffer[written,remainder])
      remainder -= nwrite
      written += nwrite
    end
    @wbuffer[0,len] = ""
  end

  public

  def write(s)
    @wbuffer = "" unless @wbuffer
    @wbuffer << s
    len = @wbuffer.size
    if @sync
      if idx = @wbuffer.rindex($/)
        len = idx+1
      else
        len = (len/BLOCK_SIZE)*BLOCK_SIZE
      end
    end
    consume_wbuf(len)
    @wbuffer[0,len] = ""
    s.size
  end

  def puts(*args)
    s = ""
    args.each{ |arg| s << arg.to_s + $/ }
    self.write(s)
    nil
  end

  def print(*args)
    s = ""
    args.each{ |arg| s << arg.to_s }
    self.write(s)
    nil
  end

  def printf(s, *args)
    self.write(s % args)
    nil
  end

  def flush
    @wbuffer = "" unless @wbuffer
    consume_wbuf(@wbuffer.size)
  end
end
